<?php

/**
 * pint, Copyright (c) 2014, mappu
 * 
 * Permission to use, copy, modify, and/or distribute this software for any
 *  purpose with or without fee is hereby granted, provided that the above
 *  copyright notice and this permission notice appear in all copies.
 */

class pint {
	protected $sz = '';
	protected $pos = 0;
	public $functions = [];
	
	public function __construct($sz='') {
		$this->setCode($sz);
		$this->functions['max'] = 'max';
	}
	
	public function setCode($sz) {
		$this->sz = $sz;
		$this->pos = 0;
	}
	
	public function nextToken($peek=false) {
		if ($this->pos >= strlen($this->sz)) return null;
		
		$matches = [];
		if (preg_match('~^\\s*\\d+(?:\\.\\d+)?~', substr($this->sz, $this->pos), $matches)) {
			$ret = floatval(ltrim($matches[0]));
			
		} elseif (preg_match('~^\\s*([a-zA-Z]+)\\s*\\(~', substr($this->sz, $this->pos), $matches)) {
			$ret = ltrim($matches[1]).'(';
			
		} elseif (preg_match('~^\\s*(?:==|\\+|\\-|\\*|\\/|\\^|\\(|[^\\s\\d]+)~', substr($this->sz, $this->pos), $matches)) {
			$ret = ltrim($matches[0]);
			
		} else {
			return null;
		}
		
		if (! $peek) {
			$this->pos += strlen($matches[0]);
		}
		
		return $ret;
	}
	
	protected function isOperator($sz) {
		switch ($sz) {
			case '==':return [0, function($a, $b) { return ($a == $b) ? 1 : 0; }];
			case '+': return [1, function($a, $b) { return $a + $b; }];
			case '-': return [1, function($a, $b) { return $a - $b; }];
			case '*': return [2, function($a, $b) { return $a * $b; }];
			case '/': return [2, function($a, $b) { return $a / $b; }];
			case '^': return [3, function($a, $b) { return pow($a, $b); }];
			default: 
				return null;
		}
	}
	
	protected function isFunction($sz) {
		$key = strtolower(substr($sz, 0, -1)); // strip (
		return array_key_exists($key, $this->functions) ? $this->functions[$key] : null;
	}
	
	protected function eval_next_atom() {
		$token = $this->nextToken(); // consume
		if (($f = $this->isFunction($token)) !== null) {
			$args = [];
			if ($this->nextToken(true) != ')') for(;;) { // check no args
				$args[] = $this->eval_next_expr();
				
				$peek = $this->nextToken(); // consume
				if ($peek == ')') break;
				if ($peek == ',') continue;
				throw new Exception('unexpected character in function call');
			}
			return call_user_func_array($f, $args); // call function
			
		} elseif ($token == '(') {
			$val = $this->eval_next_expr();
			if ($this->nextToken() != ')') { // consume
				throw new Exception('mismatched parenthesis');
			}
			return $val;
			
		} elseif ($token === null) {
			throw new Exception('unexpected end of stream');
		} elseif (($o = $this->isOperator($token)) !== null) {
			throw new Exception('unexpected binary operator');
		} else return $token;
	}
	
	protected function eval_next_expr($min_precedence=-1) {
		$lhs = $this->eval_next_atom();
		for(;;) {
			$op  = $this->isOperator($this->nextToken(true)); // peek
			if ($op === null || $op[0] < $min_precedence) break;
			
			$this->nextToken(); // consume peeked
			$lhs = $op[1]($lhs, $this->eval_next_expr($op[0] + 1));
		}
		return $lhs;
	}
	
	public static function exec($src) {
		$p = new Pint($src);
		return $p->eval_next_expr();
	}
}
